import pytest

from allure_commons.utils import now
from allure_commons.model2 import Label
from allure_commons.model2 import Status
from allure_commons.types import LabelType, AttachmentType
from allure_commons.utils import platform_label
from allure_commons.utils import host_tag, thread_tag
from allure_commons.utils import md5

from .steps import get_step_uuid
from .steps import process_gherkin_step_args
from .steps import report_remaining_steps
from .steps import report_undefined_step
from .steps import start_gherkin_step
from .steps import stop_gherkin_step
from .storage import save_excinfo
from .storage import save_test_data
from .utils import attach_data
from .utils import get_allure_description
from .utils import get_allure_description_html
from .utils import get_allure_labels
from .utils import get_allure_links
from .utils import convert_params
from .utils import get_full_name
from .utils import get_outline_params
from .utils import get_pytest_params
from .utils import get_pytest_report_status
from .utils import get_scenario_status_details
from .utils import get_test_name
from .utils import get_uuid
from .utils import post_process_test_result

from functools import partial


class PytestBDDListener:
    def __init__(self, lifecycle):
        self.lifecycle = lifecycle
        self.host = host_tag()
        self.thread = thread_tag()

    @pytest.hookimpl
    def pytest_bdd_before_scenario(self, request, feature, scenario):
        item = request.node
        uuid = get_uuid(item.nodeid)

        outline_params = get_outline_params(item)
        pytest_params = get_pytest_params(item)
        params = {**pytest_params, **outline_params}

        save_test_data(
            item=item,
            feature=feature,
            scenario=scenario,
            params=params,
        )

        full_name = get_full_name(feature, scenario)
        with self.lifecycle.schedule_test_case(uuid=uuid) as test_result:
            test_result.fullName = full_name
            test_result.name = get_test_name(item, scenario, params)
            test_result.description = get_allure_description(item, feature, scenario)
            test_result.descriptionHtml = get_allure_description_html(item)
            test_result.start = now()
            test_result.testCaseId = md5(full_name)
            test_result.labels.append(Label(name=LabelType.HOST, value=self.host))
            test_result.labels.append(Label(name=LabelType.THREAD, value=self.thread))
            test_result.labels.append(Label(name=LabelType.FRAMEWORK, value="pytest-bdd"))
            test_result.labels.append(Label(name=LabelType.LANGUAGE, value=platform_label()))
            test_result.labels.extend(get_allure_labels(item))
            test_result.links.extend(get_allure_links(item))
            test_result.parameters.extend(convert_params(outline_params, pytest_params))

        finalizer = partial(report_remaining_steps, self.lifecycle, item)
        item.addfinalizer(finalizer)

    @pytest.hookimpl
    def pytest_bdd_after_scenario(self, request, feature, scenario):
        uuid = get_uuid(request.node.nodeid)
        with self.lifecycle.update_test_case(uuid=uuid) as test_result:
            test_result.stop = now()

    @pytest.hookimpl
    def pytest_bdd_before_step(self, request, feature, scenario, step, step_func):
        start_gherkin_step(self.lifecycle, request.node, step, step_func)

    @pytest.hookimpl
    def pytest_bdd_before_step_call(self, request, feature, scenario, step, step_func, step_func_args):
        process_gherkin_step_args(self.lifecycle, request.node, step, step_func, step_func_args)

    @pytest.hookimpl
    def pytest_bdd_after_step(self, request, feature, scenario, step, step_func, step_func_args):
        stop_gherkin_step(self.lifecycle, request.node, get_step_uuid(step))

    @pytest.hookimpl
    def pytest_bdd_step_error(self, request, feature, scenario, step, step_func, step_func_args, exception):
        stop_gherkin_step(self.lifecycle, request.node, get_step_uuid(step), exception=exception)

    @pytest.hookimpl
    def pytest_bdd_step_func_lookup_error(self, request, feature, scenario, step, exception):
        report_undefined_step(self.lifecycle, request.node, step, exception)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_makereport(self, item, call):
        report = (yield).get_result()

        excinfo = call.excinfo

        status = get_pytest_report_status(report, excinfo)
        status_details = get_scenario_status_details(report, excinfo)

        uuid = get_uuid(report.nodeid)
        with self.lifecycle.update_test_case(uuid=uuid) as test_result:

            if test_result and report.when == "setup":
                test_result.status = status
                test_result.statusDetails = status_details

            if report.when == "call" and test_result:

                # Save the exception to access it from the finalizer to report
                # the remaining steps
                save_excinfo(item, excinfo)

                if test_result.status is None or test_result.status == Status.PASSED:
                    test_result.status = status
                    test_result.statusDetails = status_details

            if report.when == "teardown" and test_result:
                if test_result.status == Status.PASSED and status in [Status.FAILED, Status.BROKEN]:
                    test_result.status = status
                    test_result.statusDetails = status_details
                if report.caplog:
                    attach_data(self.lifecycle, report.caplog, "log", AttachmentType.TEXT, None)
                if report.capstdout:
                    attach_data(self.lifecycle, report.capstdout, "stdout", AttachmentType.TEXT, None)
                if report.capstderr:
                    attach_data(self.lifecycle, report.capstderr, "stderr", AttachmentType.TEXT, None)
                post_process_test_result(item, test_result)

        if report.when == 'teardown':
            self.lifecycle.write_test_case(uuid=uuid)



import csv
import io
import os
from urllib.parse import urlparse
from uuid import UUID

import pytest

from allure_commons.model2 import Label
from allure_commons.model2 import Link
from allure_commons.model2 import StatusDetails
from allure_commons.model2 import Status
from allure_commons.model2 import Parameter
from allure_commons.types import LabelType
from allure_commons.types import LinkType

from allure_commons.utils import format_exception
from allure_commons.utils import format_traceback
from allure_commons.utils import md5
from allure_commons.utils import represent
from allure_commons.utils import SafeFormatter
from allure_commons.utils import uuid4

from .storage import get_test_data

ALLURE_TITLE_ATTR = "__allure_display_name__"
ALLURE_DESCRIPTION_MARK = "allure_description"
ALLURE_DESCRIPTION_HTML_MARK = "allure_description_html"
ALLURE_LABEL_MARK = 'allure_label'
ALLURE_LINK_MARK = 'allure_link'

MARK_NAMES_TO_IGNORE = {
    "usefixtures",
    "filterwarnings",
    "skip",
    "skipif",
    "xfail",
    "parametrize",
}


def get_allure_title_of_test(item, params):
    obj = getattr(item, "obj", None)
    if obj is not None:
        return get_allure_title(obj, params)


def get_allure_title(fn, kwargs):
    if fn is not None:
        title_format = getattr(fn, ALLURE_TITLE_ATTR, None)
        if title_format:
            return interpolate(title_format, kwargs)


def interpolate(format_str, kwargs):
    return SafeFormatter().format(format_str, **kwargs) if kwargs else format_str


def get_allure_description(item, feature, scenario):
    value = get_marker_value(item, ALLURE_DESCRIPTION_MARK)
    if value:
        return value

    feature_description = extract_description(feature)
    scenario_description = extract_description(scenario)
    return "\n\n".join(filter(None, [feature_description, scenario_description]))


def get_allure_description_html(item):
    return get_marker_value(item, ALLURE_DESCRIPTION_HTML_MARK)


def iter_all_labels(item):
    for mark in item.iter_markers(name=ALLURE_LABEL_MARK):
        name = mark.kwargs.get("label_type")
        if name:
            yield from ((name, value) for value in mark.args or [])


def iter_label_values(item, name):
    return (pair for pair in iter_all_labels(item) if pair[0] == name)


def convert_labels(labels):
    return [Label(name, value) for name, value in labels]


def get_allure_labels(item):
    return convert_labels(iter_all_labels(item))


def iter_all_links(item):
    for marker in item.iter_markers(name=ALLURE_LINK_MARK):
        url = marker.args[0] if marker and marker.args else None
        if url:
            yield url, marker.kwargs.get("name"), marker.kwargs.get("link_type")


def convert_links(links):
    return [Link(url=url, name=name, type=link_type) for url, name, link_type in links]


def get_allure_links(item):
    return convert_links(iter_all_links(item))


def get_link_patterns(config):
    patterns = {}
    for link_type, pattern in config.option.allure_link_pattern:
        patterns[link_type] = pattern
    return patterns


def is_url(maybeUrl):
    try:
        result = urlparse(maybeUrl)
    except AttributeError:
        return False

    return result and (
        getattr(result, "scheme", None) or getattr(result, "netloc", None)
    )


def apply_link_pattern(patterns, link_type, url):
    if is_url(url):
        return url

    pattern = patterns.get(link_type or LinkType.LINK)
    return url if pattern is None else pattern.format(url)


def get_marker_value(item, keyword):
    marker = item.get_closest_marker(keyword)
    return marker.args[0] if marker and marker.args else None


def should_convert_mark_to_tag(mark):
    return mark.name not in MARK_NAMES_TO_IGNORE and\
        not mark.args and not mark.kwargs


def iter_pytest_tags(item):
    for mark in item.iter_markers():
        if should_convert_mark_to_tag(mark):
            yield LabelType.TAG, mark.name


def extract_description(obj):
    description = getattr(obj, "description", None)

    if isinstance(description, str):
        return description

    if not isinstance(description, list):
        return None

    while description and description[0] == "":
        description = description[1:]
    while description and description[-1] == "":
        description = description[:-1]
    return "\n".join(description) or None


def get_test_name(node, scenario, params):
    return get_allure_title_of_test(node, params) or scenario.name


def get_full_name(feature, scenario):
    feature_path = os.path.normpath(feature.rel_filename)
    return f"{feature_path}:{scenario.name}"


def get_uuid(*args):
    return str(UUID(md5(*args)))


def get_status(exception):
    if exception:
        if isinstance(exception, (pytest.skip.Exception, pytest.xfail.Exception)):
            return Status.SKIPPED
        elif isinstance(exception, (AssertionError, pytest.fail.Exception)):
            return Status.FAILED
        return Status.BROKEN
    else:
        return Status.PASSED


def get_status_details(exception, exception_type=None, traceback=None):
    if exception_type is None and exception is not None:
        exception_type = type(exception)
    message = format_exception(exception_type, exception)
    trace = format_traceback(traceback or getattr(exception, "__traceback__", None))
    return StatusDetails(message=message, trace=trace) if message or trace else None


def get_pytest_report_status(pytest_report, excinfo):
    if pytest_report.failed:
        return get_status(excinfo.value) if excinfo else Status.BROKEN

    if pytest_report.passed:
        return Status.PASSED

    if pytest_report.skipped:
        return Status.SKIPPED


def is_runtime_xfail(excinfo):
    return isinstance(excinfo.value, pytest.xfail.Exception)


def get_scenario_status_details(report, excinfo):
    if excinfo:
        message = excinfo.exconly()
        trace = report.longreprtext
        if not is_runtime_xfail(excinfo) and hasattr(report, "wasxfail"):
            reason = report.wasxfail
            message = (f"XFAIL {reason}" if reason else "XFAIL") + "\n\n" + message
        return StatusDetails(message=message, trace=trace)
    elif report.passed and hasattr(report, "wasxfail"):
        reason = report.wasxfail
        return StatusDetails(message=f"XPASS {reason}" if reason else "XPASS")
    elif report.failed and "XPASS(strict)" in report.longrepr:
        return StatusDetails(message=report.longrepr)


def get_outline_params(node):
    if hasattr(node, 'callspec'):
        return node.callspec.params.get('_pytest_bdd_example', {})
    return {}


def get_pytest_params(node):
    if hasattr(node, 'callspec'):
        pytest_params = dict(node.callspec.params)
        if "_pytest_bdd_example" in pytest_params:
            del pytest_params["_pytest_bdd_example"]
        return pytest_params
    return {}


def convert_params(outline_params, pytest_params):
    return [
        *(Parameter(
            name=name,
            value=value,
        ) for name, value in outline_params.items()),
        *(Parameter(
            name=name,
            value=represent(value),
        ) for name, value in pytest_params.items() if name not in outline_params),
    ]


def iter_pytest_labels(item, test_result):
    test_data = get_test_data(item)

    existing_labels = {label.name for label in test_result.labels}

    if LabelType.FEATURE not in existing_labels:
        yield LabelType.FEATURE, test_data.feature.name

    yield from iter_pytest_tags(item)


def iter_default_labels(item, test_result):
    return (
        Label(
            name=name,
            value=value,
        ) for name, value in iter_pytest_labels(item, test_result)
    )


def get_history_id(test_case_id, parameters, pytest_params):
    parameters_part = md5(*(pytest_params.get(p.name, p.value) for p in sorted(
        filter(lambda p: not p.excluded, parameters),
        key=lambda p: p.name,
    )))
    return f"{test_case_id}.{parameters_part}"


def post_process_test_result(item, test_result):
    test_data = get_test_data(item)

    test_result.labels.extend(iter_default_labels(item, test_result))
    test_result.historyId = get_history_id(
        test_case_id=test_result.testCaseId,
        parameters=test_result.parameters,
        pytest_params=test_data.params,
    )


def attach_data(lifecycle, body, name, attachment_type, extension=None, parent_uuid=None):
    lifecycle.attach_data(
        uuid4(),
        body,
        name=name,
        attachment_type=attachment_type,
        extension=extension,
        parent_uuid=parent_uuid,
    )


def attach_file(lifecycle, source, name, attachment_type, extension=None):
    lifecycle.attach_file(
        uuid4(),
        source,
        name=name,
        attachment_type=attachment_type,
        extension=extension,
    )


def format_csv(rows):
    with io.StringIO() as buffer:
        writer = csv.writer(buffer)
        writer.writerow(rows[0])
        writer.writerows(rows[1:])
        return buffer.getvalue()



import pytest

import allure_commons

from allure_commons.model2 import Label
from allure_commons.model2 import Link
from allure_commons.model2 import Parameter
from allure_commons.utils import represent

from .utils import ALLURE_DESCRIPTION_HTML_MARK
from .utils import ALLURE_DESCRIPTION_MARK
from .utils import ALLURE_LABEL_MARK
from .utils import ALLURE_LINK_MARK
from .utils import ALLURE_TITLE_ATTR

from .utils import apply_link_pattern
from .utils import attach_data
from .utils import attach_file
from .utils import get_link_patterns
from .steps import start_step
from .steps import stop_step


class AllurePytestBddApiHooks:
    def __init__(self, config, lifecycle):
        self.lifecycle = lifecycle
        self.__link_patterns = get_link_patterns(config)

    @allure_commons.hookimpl
    def decorate_as_title(self, test_title):

        def decorator(fn):
            setattr(fn, ALLURE_TITLE_ATTR, test_title)
            return fn

        return decorator

    @allure_commons.hookimpl
    def add_title(self, test_title):
        with self.lifecycle.update_test_case() as test_result:
            test_result.name = test_title

    @allure_commons.hookimpl
    def decorate_as_description(self, test_description):
        allure_description_mark = getattr(pytest.mark, ALLURE_DESCRIPTION_MARK)
        return allure_description_mark(test_description)

    @allure_commons.hookimpl
    def add_description(self, test_description):
        with self.lifecycle.update_test_case() as test_result:
            test_result.description = test_description

    @allure_commons.hookimpl
    def decorate_as_description_html(self, test_description_html):
        allure_description_html_mark = getattr(pytest.mark, ALLURE_DESCRIPTION_HTML_MARK)
        return allure_description_html_mark(test_description_html)

    @allure_commons.hookimpl
    def add_description_html(self, test_description_html):
        with self.lifecycle.update_test_case() as test_result:
            test_result.descriptionHtml = test_description_html

    @allure_commons.hookimpl
    def decorate_as_label(self, label_type, labels):
        allure_label_mark = getattr(pytest.mark, ALLURE_LABEL_MARK)
        return allure_label_mark(*labels, label_type=label_type)

    @allure_commons.hookimpl
    def add_label(self, label_type, labels):
        with self.lifecycle.update_test_case() as test_result:
            test_result.labels.extend(
                Label(name=label_type, value=value) for value in labels or []
            )

    @allure_commons.hookimpl
    def decorate_as_link(self, url, link_type, name):
        url = apply_link_pattern(self.__link_patterns, link_type, url)
        allure_link_mark = getattr(pytest.mark, ALLURE_LINK_MARK)
        return allure_link_mark(url, name=name, link_type=link_type)

    @allure_commons.hookimpl
    def add_link(self, url, link_type, name):
        url = apply_link_pattern(self.__link_patterns, link_type, url)
        with self.lifecycle.update_test_case() as test_result:
            test_result.links.append(Link(url=url, name=name, type=link_type))

    @allure_commons.hookimpl
    def add_parameter(self, name, value, excluded, mode):
        with self.lifecycle.update_test_case() as test_result:
            test_result.parameters.append(
                Parameter(
                    name=name,
                    value=represent(value),
                    excluded=excluded,
                    mode=mode.value if mode else None,
                ),
            )

    @allure_commons.hookimpl
    def start_step(self, uuid, title, params):
        start_step(self.lifecycle, step_uuid=uuid, title=title, params=params)

    @allure_commons.hookimpl
    def stop_step(self, uuid, exc_type, exc_val, exc_tb):
        stop_step(
            self.lifecycle,
            uuid,
            exception=exc_val,
            exception_type=exc_type,
            traceback=exc_tb,
        )

    @allure_commons.hookimpl
    def attach_data(self, body, name, attachment_type, extension):
        attach_data(self.lifecycle, body, name, attachment_type, extension)

    @allure_commons.hookimpl
    def attach_file(self, source, name, attachment_type, extension):
        attach_file(self.lifecycle, source, name, attachment_type, extension)



import pytest


ALLURE_PYTEST_BDD_HASHKEY = pytest.StashKey()


class AllurePytestBddTestData:

    def __init__(self, feature, scenario, params):
        self.feature = feature
        self.scenario = scenario
        self.params = params
        self.excinfo = None
        self.reported_steps = set()


def save_test_data(item, feature, scenario, params):
    item.stash[ALLURE_PYTEST_BDD_HASHKEY] = AllurePytestBddTestData(
        feature=feature,
        scenario=scenario,
        params=params,
    )


def get_test_data(item):
    return item.stash.get(ALLURE_PYTEST_BDD_HASHKEY, (None, None))


def get_saved_params(item):
    return get_test_data(item).params


def save_excinfo(item, excinfo):
    test_data = get_test_data(item)
    if test_data:
        test_data.excinfo = excinfo


def save_reported_step(item, step_uuid):
    test_data = get_test_data(item)
    if test_data:
        test_data.reported_steps.add(step_uuid)



from allure import attachment_type
from allure_commons.model2 import StatusDetails
from allure_commons.model2 import Status
from allure_commons.model2 import Parameter
from allure_commons.utils import format_exception
from allure_commons.utils import represent

from .storage import get_saved_params
from .storage import get_test_data
from .storage import save_reported_step
from .utils import attach_data
from .utils import format_csv
from .utils import get_allure_title
from .utils import get_uuid
from .utils import get_status
from .utils import get_status_details


def get_step_name(item, step, step_func, step_func_args=None):
    return get_allure_title_of_step(item, step_func, step_func_args) or \
        f"{step.keyword} {step.name}"


def get_allure_title_of_step(item, step_func, step_func_args):
    return get_allure_title(
        step_func,
        {
            **(get_saved_params(item) or {}),
            **(step_func_args or {}),
        },
    )


def get_step_uuid(step):
    return get_uuid(str(id(step)))


def start_step(lifecycle, step_uuid, title, params=None, parent_uuid=None):
    with lifecycle.start_step(uuid=step_uuid, parent_uuid=parent_uuid) as step_result:
        step_result.name = title
        if params:
            step_result.parameters.extend(
                Parameter(
                    name=name,
                    value=represent(value),
                ) for name, value in params.items()
            )


def stop_step(lifecycle, uuid, status=None, status_details=None, exception=None, exception_type=None, traceback=None):
    with lifecycle.update_step(uuid=uuid) as step_result:
        if step_result is None:
            return False
        step_result.status = status or get_status(exception)
        step_result.statusDetails = status_details or get_status_details(exception, exception_type, traceback)
    lifecycle.stop_step(uuid=uuid)
    return True


def start_gherkin_step(lifecycle, item, step, step_func=None, step_uuid=None):
    if step_uuid is None:
        step_uuid = get_step_uuid(step)

    start_step(
        lifecycle,
        step_uuid=step_uuid,
        title=get_step_name(item, step, step_func),
        parent_uuid=get_uuid(item.nodeid),
    )


def process_gherkin_step_args(lifecycle, item, step, step_func, step_func_args):
    allure_step_params = dict(step_func_args)
    step_uuid = get_step_uuid(step)

    docstring = step_func_args.get("docstring")
    if try_attach_docstring(lifecycle, step_uuid, docstring):
        del allure_step_params["docstring"]

    datatable = step_func_args.get("datatable")
    if try_attach_datatable(lifecycle, step_uuid, datatable):
        del allure_step_params["datatable"]

    add_step_parameters(lifecycle, step_uuid, allure_step_params)

    update_step_name(lifecycle, item, step_uuid, step_func, step_func_args)


def try_attach_docstring(lifecycle, step_uuid, docstring):
    if isinstance(docstring, str):
        attach_data(
            lifecycle=lifecycle,
            body=docstring,
            name="Doc string",
            attachment_type=attachment_type.TEXT,
            parent_uuid=step_uuid,
        )
        return True
    return False


def try_attach_datatable(lifecycle, step_uuid, datatable):
    if is_datatable(datatable):
        attach_data(
            lifecycle=lifecycle,
            body=format_csv(datatable),
            name="Data table",
            attachment_type=attachment_type.CSV,
            parent_uuid=step_uuid,
        )
        return True
    return False


def add_step_parameters(lifecycle, step_uuid, step_params):
    if not step_params:
        return

    with lifecycle.update_step(uuid=step_uuid) as step_result:
        if step_result is None:
            return

        step_result.parameters.extend(
            Parameter(
                name=name,
                value=represent(value),
            ) for name, value in step_params.items()
        )


def update_step_name(lifecycle, item, step_uuid, step_func, step_func_args):
    if not step_func_args:
        return

    new_name = get_allure_title_of_step(item, step_func, step_func_args)
    if new_name is None:
        return

    with lifecycle.update_step(uuid=step_uuid) as step_result:
        if step_result is not None:
            step_result.name = new_name


def is_datatable(value):
    return isinstance(value, list) and all(isinstance(row, list) for row in value)


def stop_gherkin_step(lifecycle, item, step_uuid, **kwargs):
    res = stop_step(lifecycle, step_uuid, **kwargs)
    if res:
        save_reported_step(item, step_uuid)
    return res


def ensure_gherkin_step_reported(lifecycle, item, step, step_uuid=None, **kwargs):

    if not step_uuid:
        step_uuid = get_step_uuid(step)

    if stop_gherkin_step(lifecycle, item, step_uuid, **kwargs):
        return

    start_gherkin_step(lifecycle, item, step, step_uuid=step_uuid)
    stop_gherkin_step(lifecycle, item, step_uuid, **kwargs)


def report_undefined_step(lifecycle, item, step, exception):
    ensure_gherkin_step_reported(
        lifecycle,
        item,
        step,
        status=Status.BROKEN,
        status_details=StatusDetails(
            message=format_exception(type(exception), exception),
        ),
    )


def report_remaining_steps(lifecycle, item):
    test_data = get_test_data(item)
    scenario = test_data.scenario
    excinfo = test_data.excinfo
    reported_steps = test_data.reported_steps

    for step in scenario.steps:
        step_uuid = get_step_uuid(step)
        if step_uuid not in reported_steps:
            __report_remaining_step(lifecycle, item, step, step_uuid, excinfo)
            excinfo = None  # Only show the full message and traceback once


def __report_remaining_step(lifecycle, item, step, step_uuid, excinfo):
    args = [lifecycle, item, step, step_uuid]
    kwargs = {
        "exception": excinfo.value,
        "exception_type": excinfo.type,
        "traceback": excinfo.tb,
    } if __is_step_running(lifecycle, step_uuid) and excinfo else {"status": Status.SKIPPED}

    ensure_gherkin_step_reported(*args, **kwargs)


def __is_step_running(lifecycle, step_uuid):
    with lifecycle.update_step(uuid=step_uuid) as step_result:
        return step_result is not None



import argparse
import os

import allure_commons
from allure_commons.logger import AllureFileLogger
from allure_commons.lifecycle import AllureLifecycle

from .allure_api_listener import AllurePytestBddApiHooks
from .pytest_bdd_listener import PytestBDDListener

from .utils import ALLURE_DESCRIPTION_MARK
from .utils import ALLURE_DESCRIPTION_HTML_MARK
from .utils import ALLURE_LABEL_MARK
from .utils import ALLURE_LINK_MARK


def pytest_addoption(parser):
    parser.getgroup("reporting").addoption('--alluredir',
                                           action="store",
                                           dest="allure_report_dir",
                                           metavar="DIR",
                                           default=None,
                                           help="Generate Allure report in the specified directory (may not exist)")

    parser.getgroup("reporting").addoption('--clean-alluredir',
                                           action="store_true",
                                           dest="clean_alluredir",
                                           help="Clean alluredir folder if it exists")

    def link_pattern(string):
        pattern = string.split(':', 1)
        if not pattern[0]:
            raise argparse.ArgumentTypeError("A link type is mandatory")

        if len(pattern) != 2:
            raise argparse.ArgumentTypeError("A link pattern is mandatory")
        return pattern

    parser.getgroup("general").addoption(
        "--allure-link-pattern",
        action="append",
        dest="allure_link_pattern",
        metavar="LINK_TYPE:LINK_PATTERN",
        default=[],
        type=link_pattern,
        help="""A URL pattern for a link type. Allows short links in tests,
        e.g., 'issue-1'. `pattern.format(short_url)` will be called to get
        the full URL"""
    )


def cleanup_factory(plugin):
    def clean_up():
        name = allure_commons.plugin_manager.get_name(plugin)
        allure_commons.plugin_manager.unregister(name=name)
    return clean_up


def register_marks(config):
    config.addinivalue_line("markers", f"{ALLURE_DESCRIPTION_MARK}: allure description")
    config.addinivalue_line("markers", f"{ALLURE_DESCRIPTION_HTML_MARK}: allure description in HTML")
    config.addinivalue_line("markers", f"{ALLURE_LABEL_MARK}: allure label marker")
    config.addinivalue_line("markers", f"{ALLURE_LINK_MARK}: allure link marker")


def pytest_configure(config):
    register_marks(config)

    report_dir = config.option.allure_report_dir
    clean = False if config.option.collectonly else config.option.clean_alluredir

    if report_dir:
        report_dir = os.path.abspath(report_dir)

        lifecycle = AllureLifecycle()

        pytest_bdd_listener = PytestBDDListener(lifecycle)
        config.pluginmanager.register(pytest_bdd_listener)
        allure_commons.plugin_manager.register(pytest_bdd_listener)
        config.add_cleanup(cleanup_factory(pytest_bdd_listener))

        allure_api_impl = AllurePytestBddApiHooks(config, lifecycle)
        allure_commons.plugin_manager.register(allure_api_impl)
        config.add_cleanup(cleanup_factory(allure_api_impl))

        file_logger = AllureFileLogger(report_dir, clean)
        allure_commons.plugin_manager.register(file_logger)
        config.add_cleanup(cleanup_factory(file_logger))
